/*******************************************************************************
 * Copyright (c) 2012, 2013 E.D.Willink and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     E.D.Willink - initial API and implementation
 *******************************************************************************/
package org.eclipse.ocl.examples.test.xtext;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.tools.DiagnosticCollector;
import javax.tools.JavaCompiler;
import javax.tools.JavaCompiler.CompilationTask;
import javax.tools.JavaFileObject;
import javax.tools.SimpleJavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;

import org.apache.log4j.Logger;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.URIUtil;
import org.eclipse.emf.codegen.ecore.generator.Generator;
import org.eclipse.emf.codegen.ecore.generator.GeneratorAdapterFactory;
import org.eclipse.emf.codegen.ecore.genmodel.GenJDKLevel;
import org.eclipse.emf.codegen.ecore.genmodel.GenModel;
import org.eclipse.emf.codegen.ecore.genmodel.generator.GenBaseGeneratorAdapter;
import org.eclipse.emf.codegen.ecore.genmodel.generator.GenModelGeneratorAdapterFactory;
import org.eclipse.emf.codegen.ecore.genmodel.util.GenModelUtil;
import org.eclipse.emf.common.EMFPlugin;
import org.eclipse.emf.common.util.BasicMonitor;
import org.eclipse.emf.common.util.Diagnostic;
import org.eclipse.emf.common.util.Monitor;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.EcoreResourceFactoryImpl;
import org.eclipse.emf.mwe.core.ConfigurationException;
import org.eclipse.jdt.annotation.NonNull;
import org.eclipse.jdt.annotation.Nullable;
import org.eclipse.ocl.examples.codegen.oclinecore.OCLinEcoreGeneratorAdapterFactory;
import org.eclipse.ocl.examples.domain.utilities.StandaloneProjectMap;
import org.eclipse.ocl.examples.domain.values.util.ValuesUtil;
import org.eclipse.ocl.examples.library.oclstdlib.OCLstdlibPackage;
import org.eclipse.ocl.examples.pivot.Type;
import org.eclipse.ocl.examples.pivot.helper.OCLHelper;
import org.eclipse.ocl.examples.pivot.manager.MetaModelManager;
import org.eclipse.ocl.examples.pivot.tests.PivotTestSuite;
import org.eclipse.ocl.examples.pivot.utilities.PivotUtil;
import org.eclipse.xtext.diagnostics.ExceptionDiagnostic;
import org.osgi.framework.Bundle;

/**
 * Tests that load a model and verify that there are no unresolved proxies as a
 * result.
 */
public class UsageTests
		extends PivotTestSuite// XtextTestCase
{

	private static final class JavaSourceFileObject
			extends SimpleJavaFileObject {

		private JavaSourceFileObject(java.net.URI uri) {
			super(uri, Kind.SOURCE);
		}

		@Override
		public CharSequence getCharContent(boolean ignoreEncodingErrors) throws IOException {
			char[] buf = new char[4096];
			StringBuffer s = new StringBuffer();
			Reader reader = new FileReader(new File(uri));
			try {
				int len;
				while ((len = reader.read(buf)) > 0) {
					s.append(buf, 0, len);
				}
			}
			finally {
				reader.close();
			}
			return s.toString();
		}
	}

	public Logger log;

	/**
	 * Checks all resources in a resource set for any errors or warnings.
	 * 
	 * @param resourceSet
	 * @throws ConfigurationException
	 *             if any error present
	 */
	public void checkResourceSet(ResourceSet resourceSet)
			throws ConfigurationException {
		int errorCount = 0;
		for (Resource aResource : resourceSet.getResources()) {
			List<Resource.Diagnostic> errors = aResource.getErrors();
			if (errors.size() > 0) {
				for (Resource.Diagnostic error : errors) {
					if (error instanceof ExceptionDiagnostic) {
						log.error("Error for '" + aResource.getURI() + "'",
							((ExceptionDiagnostic) error).getException());
					} else {
						log.error(error + " for '" + aResource.getURI() + "'");
					}
					errorCount++;
				}
			}
			List<Resource.Diagnostic> warnings = aResource.getWarnings();
			if (warnings.size() > 0) {
				for (Resource.Diagnostic warning : warnings) {
					if (warning instanceof ExceptionDiagnostic) {
						log.warn("Warning for '" + aResource.getURI() + "'",
							((ExceptionDiagnostic) warning).getException());
					} else {
						log.warn(warning + " for '" + aResource.getURI() + "'");
					}
				}
			}
		}
		if (errorCount > 0) {
			throw new RuntimeException("Errors in ResourceSet");
		}
	}

	protected void getCompilationUnits(@NonNull List<JavaFileObject> compilationUnits,
			@NonNull File directory) throws Exception {
		File[] files = directory.listFiles();
		if (files != null) {
			for (File file : files) {
				if (file.isDirectory()) {
					getCompilationUnits(compilationUnits, file);
				} else if (file.isFile()) {
//					System.out.println("Compiling " + file);
					compilationUnits.add(new JavaSourceFileObject(file.toURI()));
				}
			}
		}
	}

	protected void getCompilationUnits(@NonNull List<JavaFileObject> compilationUnits,
			@NonNull IContainer container) throws CoreException {
		for (IResource member : container.members()) {
			if (member instanceof IContainer) {
				getCompilationUnits(compilationUnits, (IContainer) member);
			} else if ((member instanceof IFile)
				&& member.getFileExtension().equals("java")) {
				java.net.URI locationURI = member.getLocationURI();
//				System.out.println("Compiling " + locationURI);
				compilationUnits.add(new JavaSourceFileObject(locationURI));
			}
		}
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
		log = Logger.getLogger(UsageTests.class);
		// AcceleoNature.class.getName(); // Pull in the plugin for Hudson
		doOCLinEcoreSetup();
		configurePlatformResources();
//		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap()
//			.put("pivot", new XMIResourceFactoryImpl()); //$NON-NLS-1$
	}

	@Override
	protected void tearDown()
			throws Exception {
		log = null;
		uninstall();
		super.tearDown();
	}

	protected @NonNull String createClassPath(@NonNull List<String> projectNames) {
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		String pathSeparator = null;
		StringBuilder s = new StringBuilder();
		for (String projectName : projectNames) {
			String projectPath = null;
			IProject project = root.getProject(projectName);
			if (project != null) {
				IPath location = project.getLocation();
				if (location != null) {
					projectPath = location.toString() + "/";
				}
			}
			if (projectPath == null) {
				Bundle bundle = Platform.getBundle(projectName);
				projectPath = bundle.getLocation();
			}
			
			if (projectPath.startsWith("reference:")) {
				projectPath = projectPath.substring(10);
			}
			URI uri = URI.createURI(projectPath);
			if (uri.isFile()) {
				projectPath =  uri.toFileString().replace("\\", "/");
			}
			if (projectPath.endsWith("/")) {
				projectPath = projectPath + "bin";
			}
			if (pathSeparator != null) {
				s.append(pathSeparator);
			}
			else {
				pathSeparator = System.getProperty("path.separator");
			}
			s.append(projectPath);
		}
		@SuppressWarnings("null")@NonNull String string = s.toString();
		return string;
	}
	
	public @NonNull String createGenModelContent(@NonNull String testProjectPath, @NonNull String fileName, @Nullable String usedGenPackages) {
		StringBuilder s = new StringBuilder();
		s.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		s.append("<genmodel:GenModel xmi:version=\"2.0\"\n");
		s.append("    xmlns:xmi=\"http://www.omg.org/XMI\"\n");
		s.append("    xmlns:ecore=\"http://www.eclipse.org/emf/2002/Ecore\"\n");
		s.append("    xmlns:genmodel=\"http://www.eclipse.org/emf/2002/GenModel\"\n");
		s.append("    modelDirectory=\"/" + testProjectPath + "/src-gen\"\n");
		s.append("    modelPluginID=\"" + fileName + "." + fileName + "\"\n");
		s.append("    modelName=\"" + fileName + "\"\n");
		s.append("    importerID=\"org.eclipse.emf.importer.ecore\"\n");
		s.append("    complianceLevel=\"6.0\"\n");
		s.append("    operationReflection=\"true\"\n");
		s.append("    copyrightFields=\"false\"\n");
		s.append("    bundleManifest=\"false\"\n");
		s.append("    usedGenPackages=\"");
		if (usedGenPackages != null) {
			s.append(usedGenPackages + " ");
		}
		s.append("platform:/resource/org.eclipse.ocl.examples.library/model/oclstdlib.genmodel#//oclstdlib\"\n");
		s.append("    updateClasspath=\"false\">\n");
		s.append("  <genAnnotations source=\"http://www.eclipse.org/OCL/GenModel\">\n");
		s.append("    <details key=\"Use Delegates\" value=\"false\"/>\n");
		s.append("    <details key=\"Use Null Annotations\" value=\"true\"/>\n");
		s.append("  </genAnnotations>\n");
		s.append("  <foreignModel>" + fileName + ".ecore</foreignModel>\n");
		s.append("  <genPackages prefix=\"" + fileName + "\"\n");
		s.append("    disposableProviderFactory=\"true\"\n");
		s.append("    ecorePackage=\"" + fileName + ".ecore#/\">\n");
		s.append("  </genPackages>\n");
		s.append("</genmodel:GenModel>\n");
		s.append("\n");;
		@SuppressWarnings("null")@NonNull String string = s.toString();
		return string;
	}

	public void createGenModelFile(String fileName, String fileContent)
			throws IOException {
		File file = new File(getProjectFile(), fileName);
		Writer writer = new FileWriter(file);
		writer.append(fileContent);
		writer.close();
	}

	protected @NonNull URI createModels(@NonNull String testProjectName, @NonNull String testFileStem,
			@NonNull String oclinecoreFile, @NonNull String genmodelFile)
			throws CoreException, IOException {
		if (EMFPlugin.IS_ECLIPSE_RUNNING) {
			suppressGitPrefixPopUp();
			IWorkspace workspace = ResourcesPlugin.getWorkspace();
			IProject project = workspace.getRoot().getProject(testProjectName);
			if (!project.exists()) {
				project.create(null);
			}
		}
		MetaModelManager metaModelManager2 = new MetaModelManager(getProjectMap());
		createEcoreFile(metaModelManager2, testFileStem, oclinecoreFile);
		createGenModelFile(testFileStem + ".genmodel", genmodelFile);
		URI fileURI = getProjectFileURI(testFileStem + ".genmodel");
		// System.out.println("Generating Ecore Model using '" + fileURI + "'");
		metaModelManager2.dispose();
		return fileURI;
	}

	protected boolean doCompile(@NonNull String testProjectName, @NonNull String... extraClasspathProjects) throws Exception {
		List<String> classpathProjects = new ArrayList<String>();
		classpathProjects.add("org.eclipse.jdt.annotation");
		classpathProjects.add("org.eclipse.emf.common");
		classpathProjects.add("org.eclipse.emf.ecore");
		classpathProjects.add("org.eclipse.ocl.examples.domain");
		classpathProjects.add("org.eclipse.ocl.examples.library");
		classpathProjects.add("org.eclipse.ocl.examples.pivot");
		classpathProjects.add("org.eclipse.osgi");
		for (String extraClasspathProject : extraClasspathProjects) {
			classpathProjects.add(extraClasspathProject);
		}
		List<String> compilationOptions = new ArrayList<String>();
		compilationOptions.add("-d");
		compilationOptions.add("bin");
		compilationOptions.add("-source");
		compilationOptions.add("1.5");
		compilationOptions.add("-target");
		compilationOptions.add("1.5");
		compilationOptions.add("-g");
		List<JavaFileObject> compilationUnits = new ArrayList<JavaFileObject>();
		Object context = null;
		if (EMFPlugin.IS_ECLIPSE_RUNNING) {
			IWorkspace workspace = ResourcesPlugin.getWorkspace();
			IProject project = workspace.getRoot().getProject(testProjectName);
			if (project != null) {
				getCompilationUnits(compilationUnits, project);
				java.net.URI locationURI = project.getLocationURI();
				String binURI = URIUtil.toUnencodedString(locationURI) + "/bin";
				URI uri = URI.createURI(binURI);
				if (uri.isFile()) {
					binURI = uri.toFileString().replace("\\", "/");
				}
				compilationOptions.set(1, binURI);
				new File(locationURI.getPath() + "/bin").mkdirs();
				compilationOptions.add("-cp");
				String path = createClassPath(classpathProjects);
				compilationOptions.add(path);
			}
			context = project;
		} else {
			File dir = new File("src-gen/" + testProjectName);
			getCompilationUnits(compilationUnits, dir);
			context = dir;
		}

		DiagnosticCollector<JavaFileObject> diagnostics = new DiagnosticCollector<JavaFileObject>();
		JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
		StandardJavaFileManager stdFileManager = compiler
			.getStandardFileManager(null, Locale.getDefault(), null);

		// System.out.printf("%6.3f getTask\n", 0.001 *
		// (System.currentTimeMillis()-base));
		CompilationTask compilerTask = compiler.getTask(null, stdFileManager,
			diagnostics, compilationOptions, null, compilationUnits);
		// System.out.printf("%6.3f call\n", 0.001 *
		// (System.currentTimeMillis()-base));
		if (!compilerTask.call()) {
			StringBuilder s2 = new StringBuilder();
			s2.append("javac");
			for (String compilationOption : compilationOptions) {
				s2.append(" ");
				s2.append(compilationOption);
			}
			for (JavaFileObject compilationUnit : compilationUnits) {
				s2.append("\n  ");
				s2.append(compilationUnit);
			}
			System.out.println(s2.toString());
			StringBuilder s = new StringBuilder();
			for (javax.tools.Diagnostic<?> diagnostic : diagnostics.getDiagnostics()) {
				s.append("\n" + diagnostic);
			}
			if (s.length() > 0) {
				throw new IOException("Failed to compile " + context + s.toString());
			}
			System.out.println("Compilation of " + context + " returned false but no diagnostics");
		}
		// System.out.printf("%6.3f close\n", 0.001 *
		// (System.currentTimeMillis()-base));
		stdFileManager.close(); // Close the file manager which re-opens automatically
		// System.out.printf("%6.3f forName\n", 0.001 *
		// (System.currentTimeMillis()-base));
		// Class<?> testClass = Class.forName(qualifiedName);
		// return testClass;

		return true;
	}

	protected void doDelete(@NonNull String testProjectName) throws Exception {
		if (EMFPlugin.IS_ECLIPSE_RUNNING) {
			suppressGitPrefixPopUp();
			IWorkspace workspace = ResourcesPlugin.getWorkspace();
			IProject project = workspace.getRoot().getProject(testProjectName);
			project.delete(true, true, null);
		} else {
			File dir = new File("src-gen/" + testProjectName);
			if (dir.exists()) {
				doDeleteDirectory(dir);
			}
		}
	}

	protected void doDeleteDirectory(File dir) throws Exception {
		File[] listFiles = dir.listFiles();
		for (File file : listFiles) {
			if (file.isFile()) {
				file.delete();
			} else if (file.isDirectory()) {
				doDeleteDirectory(file);
			}
		}
		dir.delete();
	}

	protected void doGenModel(@NonNull String testProjectName, @NonNull URI genmodelURI) throws Exception {
		if (EMFPlugin.IS_ECLIPSE_RUNNING) {
			suppressGitPrefixPopUp();
			IWorkspace workspace = ResourcesPlugin.getWorkspace();
			IProject project = workspace.getRoot().getProject(testProjectName);
			if (!project.exists()) {
				project.create(null);
			}
		}
//		MetaModelManager metaModelManager2 = new MetaModelManager();
//		metaModelManager = metaModelManager2;
//		GeneratorAdapterFactory.Descriptor.Registry.INSTANCE.addDescriptor( org.eclipse.emf.codegen.ecore.genmodel.GenModelPackage.eNS_URI, OCLinEcoreGeneratorAdapterFactory.DESCRIPTOR);
		URI fileURI = genmodelURI; //getProjectFileURI(testFileStem + ".genmodel");
		// System.out.println("Generating Ecore Model using '" + fileURI + "'");
//		metaModelManager2.dispose();
		MetaModelManager metaModelManager = new MetaModelManager(getProjectMap());
		ResourceSet resourceSet = metaModelManager.getExternalResourceSet();
		StandaloneProjectMap projectMap = metaModelManager.getProjectMap();
		projectMap.configure(resourceSet, StandaloneProjectMap.LoadFirstStrategy.INSTANCE, StandaloneProjectMap.MapToFirstConflictHandler.INSTANCE);
		resourceSet.getPackageRegistry().put(org.eclipse.emf.codegen.ecore.genmodel.GenModelPackage.eNS_URI,  org.eclipse.emf.codegen.ecore.genmodel.GenModelPackage.eINSTANCE);
		//FIXME this is needed so long as Pivot.merged.genmodel is a UML genmodel
		resourceSet.getPackageRegistry().put(org.eclipse.uml2.codegen.ecore.genmodel.GenModelPackage.eNS_URI,  org.eclipse.uml2.codegen.ecore.genmodel.GenModelPackage.eINSTANCE);
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("genmodel", new EcoreResourceFactoryImpl());
		GeneratorAdapterFactory.Descriptor.Registry.INSTANCE.addDescriptor( org.eclipse.emf.codegen.ecore.genmodel.GenModelPackage.eNS_URI, GenModelGeneratorAdapterFactory.DESCRIPTOR);
		GeneratorAdapterFactory.Descriptor.Registry.INSTANCE.addDescriptor( org.eclipse.emf.codegen.ecore.genmodel.GenModelPackage.eNS_URI, OCLinEcoreGeneratorAdapterFactory.DESCRIPTOR);
		if (resourceSet instanceof ResourceSetImpl) {
			ResourceSetImpl resourceSetImpl = (ResourceSetImpl) resourceSet;
			Map<URI, Resource> uriResourceMap = resourceSetImpl.getURIResourceMap();
			if (uriResourceMap != null) {
				uriResourceMap.clear();
			}
		}
		resourceSet.getResources().clear();
		Resource resource = resourceSet.getResource(fileURI, true);
		// EcoreUtil.resolveAll(resourceSet); -- genModel can fail if
		// proxies resolved here
		// problem arises if genmodel has an obsolete feature for a feature
		// moved up the inheritance hierarchy
		// since the proxy seems to be successfully resolved giving a double
		// feature
		checkResourceSet(resourceSet);
		EObject eObject = resource.getContents().get(0);
		if (!(eObject instanceof GenModel)) {
			throw new ConfigurationException("No GenModel found in '" + resource.getURI() + "'");
		}
		GenModel genModel = (GenModel) eObject;
		genModel.reconcile();
		checkResourceSet(resourceSet);
		// genModel.setCanGenerate(true);
		// validate();

		genModel.setValidateModel(true); // The more checks the better
		// genModel.setCodeFormatting(true); // Normalize layout
		genModel.setForceOverwrite(false); // Don't overwrite read-only files
		genModel.setCanGenerate(true);
		// genModel.setFacadeHelperClass(null); // Non-null gives JDT
		// default NPEs
		// genModel.setFacadeHelperClass(StandaloneASTFacadeHelper.class.getName());
		// // Bug 308069
		// genModel.setValidateModel(true);
		genModel.setBundleManifest(false); // New manifests should be generated manually
		genModel.setUpdateClasspath(false); // New class-paths should be generated manually
		genModel.setComplianceLevel(GenJDKLevel.JDK50_LITERAL);
		// genModel.setRootExtendsClass("org.eclipse.emf.ecore.impl.MinimalEObjectImpl$Container");
		Diagnostic diagnostic = genModel.diagnose();
		if (diagnostic.getSeverity() != Diagnostic.OK) {
			fail(diagnostic.toString());
		}

		/*
		 * JavaModelManager.getJavaModelManager().initializePreferences(); new
		 * JavaCorePreferenceInitializer().initializeDefaultPreferences();
		 * 
		 * GenJDKLevel genSDKcomplianceLevel = genModel.getComplianceLevel();
		 * String complianceLevel = JavaCore.VERSION_1_5; switch
		 * (genSDKcomplianceLevel) { case JDK60_LITERAL: complianceLevel =
		 * JavaCore.VERSION_1_6; case JDK14_LITERAL: complianceLevel =
		 * JavaCore.VERSION_1_4; default: complianceLevel =
		 * JavaCore.VERSION_1_5; } // Hashtable<?,?> defaultOptions =
		 * JavaCore.getDefaultOptions(); //
		 * JavaCore.setComplianceOptions(complianceLevel, defaultOptions); //
		 * JavaCore.setOptions(defaultOptions);
		 */

		Generator generator = GenModelUtil.createGenerator(genModel);
		Monitor monitor = new BasicMonitor();
		diagnostic = generator.generate(genModel, GenBaseGeneratorAdapter.MODEL_PROJECT_TYPE, monitor);
		if (diagnostic.getSeverity() != Diagnostic.OK) {
			String s = PivotUtil.formatDiagnostics(diagnostic, "\n");
			fail("Generation failure" + s);
		}
		metaModelManager.dispose();
	}

	public void testBug370824() throws Exception {
		String testFileStem = "Bug370824";
		String testProjectName = "bug370824";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile = "package bug370824 : bug370824 = 'http://bug370824'\n"
			+ "{\n"
			+ "    class Clase1\n"
			+ "    {\n"
			+ "        invariant : self.name.size() > 0;\n"
			+ "        attribute name : String[?] { ordered };\n"
			+ "    }\n"
			+ "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, "Bug370824", null);
		URI fileURI = createModels(testProjectName, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, fileURI);
	}

	public void testBug409650() throws Exception {
		String testFileStem = "Bug409650";
		String testProjectName = "bug409650";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile = "package bug409650 : bug409650 = 'http://bug409650'\n"
			+ "{\n"
			+ "    class Clase1\n"
			+ "    {\n"
			+ "        invariant : self.name.size() > 0;\n"
			+ "        attribute name : String[?] { ordered };\n"
			+ "        operation copy(b : Boolean) : Boolean { body: b; }\n"
			+ "        operation complement(b : Boolean) : Boolean { body: not b; }\n"
			+ "        operation myPrefixedName(s1 : String, s2 : String) : String { body: s1 + name + s2; }\n"
			+ "        operation me() : Clase1 { body: self.oclAsType(Clase1); }\n"
			+ "    }\n" + "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, "Bug409650", null);
		doDelete(testProjectName);
		URI fileURI = createModels(testProjectPath, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, fileURI);
		doCompile(testProjectName);
		if (!EMFPlugin.IS_ECLIPSE_RUNNING) { // FIXME find out how to get dynamic project onto classpath
			String qualifiedPackageName = testProjectName + "." + testFileStem + "Package";
			EPackage ePackage = doLoadPackage(qualifiedPackageName);
//			System.out.println("Loaded " + ePackage);
			EClass eClass = (EClass) ePackage.getEClassifier("Clase1");
			EStructuralFeature eStructuralFeature = eClass.getEStructuralFeature("name");
			EFactory eFactory = ePackage.getEFactoryInstance();
			//
			EObject eObject = eFactory.create(eClass);
			assertQueryTrue(eObject, "name = null");
			assertQueryTrue(eObject, "complement(true) = false");
			eObject.eSet(eStructuralFeature, "testing");
			assertQueryFalse(eObject, "name = null");
			assertQueryTrue(eObject, "name = 'testing'");
			assertQueryEquals(eObject, "XtestingY", "self.myPrefixedName('X', 'Y')");
			assertQueryEquals(eObject, eObject, "self.me()");
		}
	}

	public void testEcoreTypes412736() throws Exception {
		String testFileStem = "Bug412736";
		String testProjectName = "bug412736";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile = "import ecore : 'http://www.eclipse.org/emf/2002/Ecore#/';\n"
			+ "package bug412736 : bug412736 = 'http://bug412736'\n"
			+ "{\n"
			+ "    datatype ENumber : 'java.lang.Number' { serializable };\n"
			+ "    class EcoreDataTypes\n"
			+ "    {\n"
			+ "        attribute eBigDecimal : ecore::EBigDecimal { derived readonly volatile } { derivation: negEBigDecimal(1); }\n"
			+ "        attribute eBigInteger : ecore::EBigInteger { derived readonly volatile } { derivation: negEBigInteger(1); }\n"
			+ "        attribute eBooleanObject : ecore::EBooleanObject { derived readonly volatile } { derivation: notEBooleanObject(true); }\n"
			+ "        attribute eBoolean : ecore::EBoolean { derived readonly volatile } { derivation: notEBoolean(true); }\n"
			+ "        attribute eCharacterObject : ecore::ECharacterObject { derived readonly volatile } { derivation: negECharacterObject(1); }\n"
			+ "        attribute eChar : ecore::EChar { derived readonly volatile } { derivation: negEChar(1); }\n"
			+ "        attribute eDoubleObject : ecore::EDoubleObject { derived readonly volatile } { derivation: negEDoubleObject(1); }\n"
			+ "        attribute eDouble : ecore::EDouble { derived readonly volatile } { derivation: negEDouble(1); }\n"
			+ "        attribute eFloatObject : ecore::EFloatObject { derived readonly volatile } { derivation: negEFloatObject(1); }\n"
			+ "        attribute eFloat : ecore::EFloat { derived readonly volatile } { derivation: negEFloat(1); }\n"
			+ "        attribute eIntegerObject : ecore::EIntegerObject { derived readonly volatile } { derivation: negEIntegerObject(1); }\n"
			+ "        attribute eInt : ecore::EInt { derived readonly volatile } { derivation: negEInt(1); }\n"
			+ "        attribute eLongObject : ecore::ELongObject { derived readonly volatile } { derivation: negELongObject(1); }\n"
			+ "        attribute eLong : ecore::ELong { derived readonly volatile } { derivation: negELong(1); }\n"
//			+ "        attribute eNumber : ENumber { derived readonly volatile } { derivation: negENumber(ENumber{'1'}); }\n"
			+ "        attribute eShortObject : ecore::EShortObject { derived readonly volatile } { derivation: negEShortObject(1); }\n"
			+ "        attribute eShort : ecore::EShort { derived readonly volatile } { derivation: negEShort(1); }\n"
			+ "        attribute eString : ecore::EString { derived readonly volatile } { derivation: upCase('abc'); }\n"
			+ "        operation negEBigDecimal(b : ecore::EBigDecimal) : ecore::EBigDecimal { body: -b; }\n"
			+ "        operation negEBigInteger(b : ecore::EBigInteger) : ecore::EBigInteger { body: -b; }\n"
			+ "        operation negEChar(b : ecore::EChar) : ecore::EChar { body: -b; }\n"
			+ "        operation negECharacterObject(b : ecore::ECharacterObject) : ecore::ECharacterObject { body: -b; }\n"
			+ "        operation negEDouble(b : ecore::EDouble) : ecore::EDouble { body: -b; }\n"
			+ "        operation negEDoubleObject(b : ecore::EDoubleObject) : ecore::EDoubleObject { body: -b; }\n"
			+ "        operation negEFloat(b : ecore::EFloat) : ecore::EFloat { body: -b; }\n"
			+ "        operation negEFloatObject(b : ecore::EFloatObject) : ecore::EFloatObject { body: -b; }\n"
			+ "        operation negEInt(b : ecore::EInt) : ecore::EInt { body: -b; }\n"
			+ "        operation negEIntegerObject(b : ecore::EIntegerObject) : ecore::EIntegerObject { body: -b; }\n"
			+ "        operation negELong(b : ecore::ELong) : ecore::ELong { body: -b; }\n"
			+ "        operation negELongObject(b : ecore::ELongObject) : ecore::ELongObject { body: -b; }\n"
//			+ "        operation negENumber(b : ENumber) : ENumber { body: (-(b.oclAsType(Integer))).oclAsType(ENumber); }\n"
			+ "        operation negEShort(b : ecore::EShort) : ecore::EShort { body: -b; }\n"
			+ "        operation negEShortObject(b : ecore::EShortObject) : ecore::EShortObject { body: -b; }\n"
			+ "        operation notEBoolean(b : ecore::EBoolean) : ecore::EBoolean { body: not b; }\n"
			+ "        operation notEBooleanObject(b : ecore::EBooleanObject) : ecore::EBooleanObject { body: not b; }\n"
			+ "        operation upCase(b : ecore::EString) : ecore::EString { body: b.toUpper(); }\n"
			+ "    }\n" + "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, "Bug412736", null);
		doDelete(testProjectName);
		URI fileURI = createModels(testProjectPath, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, fileURI);
		doCompile(testProjectName);
		if (!EMFPlugin.IS_ECLIPSE_RUNNING) { // FIXME find out how to get dynamic project onto classpath
			String qualifiedPackageName = testProjectName + "." + testFileStem + "Package";
			EPackage ePackage = doLoadPackage(qualifiedPackageName);
			EClass eClass = (EClass) ePackage.getEClassifier("EcoreDataTypes");
			EFactory eFactory = ePackage.getEFactoryInstance();
			//
			EObject eObject = eFactory.create(eClass);
			assertQueryTrue(eObject, "eBigInteger = eBigDecimal");
			assertQueryTrue(eObject, "eChar = eCharacterObject");
			assertQueryTrue(eObject, "eBoolean = eBooleanObject");
			assertQueryTrue(eObject, "eDouble = eDoubleObject");
			assertQueryTrue(eObject, "eFloat = eFloatObject");
			assertQueryTrue(eObject, "eInt = eIntegerObject");
			assertQueryTrue(eObject, "eLong = eLongObject");
//			assertQueryTrue(eObject, "eNumber = eFloat");				-- waiting for BUG 370087
			assertQueryTrue(eObject, "eShort = eShortObject");
			assertQueryTrue(eObject, "eString = 'ABC'");
		}
	}

	public void testEnumTypes412685() throws Exception {
		// FIIXME next line compensates an uninstall overenthusiasm
		EPackage.Registry.INSTANCE.put(OCLstdlibPackage.eNS_URI, OCLstdlibPackage.eINSTANCE);
		String testFileStem = "Bug412685";
		String testProjectName = "bug412685";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile = "import ecore : 'http://www.eclipse.org/emf/2002/Ecore#/';\n"
			+ "package bug412685 : bug412685 = 'http://bug412685'\n"
			+ "{\n"
			+ "    enum Color { serializable } {\n"
			+ "    	literal BLACK;\n"
			+ "    	literal WHITE;\n"
			+ "    }\n"
			+ "    class EnumTypes\n"
			+ "    {\n"
			+ "        attribute eBlack : Color = 'BLACK' { readonly };\n"
			+ "        attribute eWhite : Color = 'WHITE' { readonly };\n"
			+ "        attribute eColor : Color { derived readonly volatile } { derivation: otherColor(Color::BLACK); }\n"
			+ "        operation opaqueColor(eColor : Color) : OclAny { body: eColor; }\n"
			+ "        operation otherColor(eColor : Color) : Color { body: if eColor = Color::BLACK then Color::WHITE else Color::BLACK endif; }\n"
			+ "    }\n"
			+ "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, testFileStem, null);
		doDelete(testProjectName);
		URI fileURI = createModels(testProjectPath, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, fileURI);
		doCompile(testProjectName);
		if (!EMFPlugin.IS_ECLIPSE_RUNNING) { // FIXME find out how to get dynamic project onto classpath
			String qualifiedPackageName = testProjectName + "." + testFileStem + "Package";
			EPackage ePackage = doLoadPackage(qualifiedPackageName);
			EClass eClass = (EClass) ePackage.getEClassifier("EnumTypes");
			EFactory eFactory = ePackage.getEFactoryInstance();
			//
			EObject eObject = eFactory.create(eClass);
			assertQueryTrue(eObject, "let aWhite : OclAny = opaqueColor(eWhite) in eColor = aWhite");
			assertQueryTrue(eObject, "let aWhite : OclAny = eWhite.oclAsType(OclAny) in eColor = aWhite");
			assertQueryTrue(eObject, "eColor = eWhite");
			assertQueryTrue(eObject, "eColor = Color::WHITE");
		}
	}

	public void testCSE() throws Exception {
//		CommonSubexpressionEliminator.CSE_PLACES.setState(true);
//		CommonSubexpressionEliminator.CSE_PRUNE.setState(true);
//		CommonSubexpressionEliminator.CSE_PULL_UP.setState(true);
//		CommonSubexpressionEliminator.CSE_PUSH_UP.setState(true);
//		CommonSubexpressionEliminator.CSE_REWRITE.setState(true);
		String testFileStem = "CSEs";
		String testProjectName = "cses";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile = "import ecore : 'http://www.eclipse.org/emf/2002/Ecore#/';\n"
			+ "package cses : cses = 'http://cses'\n"
			+ "{\n"
			+ "    class CSEs\n"
			+ "    {\n"
//			+ "        attribute a : ecore::EInt = '3' { readonly };\n"
			+ "        operation test(a : ecore::EInt, b : ecore::EInt, c : ecore::EInt) : ecore::EInt { body: if a + b + c > 0 then a + b + c else a + b endif; }\n"
			+ "    }\n"
			+ "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, testFileStem, null);
		doDelete(testProjectName);
		URI fileURI = createModels(testProjectPath, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, fileURI);
		doCompile(testProjectName);
		if (!EMFPlugin.IS_ECLIPSE_RUNNING) { // FIXME find out how to get dynamic project onto classpath
			String qualifiedPackageName = testProjectName + "." + testFileStem + "Package";
			EPackage ePackage = doLoadPackage(qualifiedPackageName);
			EClass eClass = (EClass) ePackage.getEClassifier("CSEs");
			EFactory eFactory = ePackage.getEFactoryInstance();
			//
			EObject eObject = eFactory.create(eClass);
			OCLHelper helper = getHelper();
			Type contextType = helper.getOCL().getMetaModelManager().getType(idResolver.getStaticTypeOf(eObject));
			helper.setContext(contextType);
//			ExpressionInOCL query = helper.createQuery("test(3, 2, 1)");
//			assertCallCount(query, null, 2);
//			assertCallCount(query, NumericPlusOperation.INSTANCE, 2);
			assertQueryEquals(eObject, 6, "test(3, 2, 1)");
			assertQueryEquals(eObject, -5, "test(3, -8, 1)");
		}
	}

/*	private void assertCallCount(ExpressionInOCL query, @Nullable LibraryOperation calledOperation, int expectedCount) {
		List<CGOperationCallExp> calls = new ArrayList<CGOperationCallExp>();
		for (TreeIterator<EObject> tit = query.getBodyExpression().eAllContents(); tit.hasNext(); ) {
			EObject eObject = tit.next();
			if (eObject instanceof CGOperationCallExp) {
				CGOperationCallExp callExp = (CGOperationCallExp)eObject;
				if (calledOperation == null) {
					calls.add(callExp);
				} else if (callExp.getReferredOperation().getImplementation() == calledOperation) {
					calls.add(callExp);
				}
			}
		}
		if (calledOperation == null) {
			assertEquals("Mismatching call count", expectedCount, calls.size());
		}
		else {
			assertEquals("Mismatching call count of " + calledOperation, expectedCount, calls.size());
		}
	} */

	public void testSysML_QUDV() throws Exception {
//		CommonSubexpressionEliminator.CSE_BUILD.setState(true);
//		CommonSubexpressionEliminator.CSE_PLACES.setState(true);
//		CommonSubexpressionEliminator.CSE_PRUNE.setState(true);
//		CommonSubexpressionEliminator.CSE_PULL_UP.setState(true);
//		CommonSubexpressionEliminator.CSE_PUSH_UP.setState(true);
//		CommonSubexpressionEliminator.CSE_REWRITE.setState(true);
//		String testFileStem = "SysML_ValueTypes_QUDV";
		String testProjectName = "SysML_ValueTypes_QUDV";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		doDelete(testProjectName);
		//		MetaModelManager metaModelManager2 = new MetaModelManager();
//		createEcoreFile(metaModelManager2, "Dummy" + testFileStem, "");
//		metaModelManager2.dispose();
		@SuppressWarnings("null")@NonNull URI genModelURI = URI.createPlatformResourceURI("/" + ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS + "/model/SysML_ValueTypes_QUDV.genmodel", true);
		if (!resourceSet.getURIConverter().exists(genModelURI, null)) {
			return;
		}			
		doGenModel(testProjectPath, genModelURI);
		doCompile(testProjectName);
		if (!EMFPlugin.IS_ECLIPSE_RUNNING) { // FIXME find out how to get dynamic project onto classpath
			String qualifiedPackageName = testProjectName + ".QUDV.QUDVPackage";
			EPackage ePackage = doLoadPackage(qualifiedPackageName);
			EClass eClass = (EClass) ePackage.getEClassifier("DerivedQuantityKind");
			EFactory eFactory = ePackage.getEFactoryInstance();
			//
			EObject eObject = eFactory.create(eClass);
			assertQueryTrue(eObject, "dependsOnQuantityKinds() <> null");
		}
		if (!EMFPlugin.IS_ECLIPSE_RUNNING) { // FIXME find out how to get dynamic project onto classpath
			String qualifiedPackageName = testProjectName + ".PrimitiveValueTypes.PrimitiveValueTypesPackage";
			EPackage ePackage = doLoadPackage(qualifiedPackageName);
			EClass eClass = (EClass) ePackage.getEClassifier("Complex");
			EFactory eFactory = ePackage.getEFactoryInstance();
			//
			EObject eObject = eFactory.create(eClass);
			assertQueryTrue(eObject, "imaginaryPart = realPart");
			assertQueryTrue(eObject, "oclType() <> null");
			assertQueryTrue(eObject, "oclType().oclIsKindOf(OclAny)");
		}
	}

	protected EPackage doLoadPackage(@NonNull String qualifiedModelPackageName) throws Exception {
		Class<?> testClass = Class.forName(qualifiedModelPackageName);
//		System.out.println("Loaded " + testClass.getName());
		Object eInstance = testClass.getDeclaredField("eINSTANCE").get(null);
		return (EPackage) eInstance;
	}

	public void testInitStatics() {
		assertTrue(ValuesUtil.initAllStatics());
		assertFalse(ValuesUtil.initAllStatics());
	}
	
	public void testPivotMetamodelImport414855() throws Exception {
		String testFileStem = "Bug414855";
		String testProjectName = "bug414855";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile =
			"import pivot : 'http://www.eclipse.org/ocl/3.1.0/Pivot#/';\n"
			+ "package bug414855 : bug414855 = 'http://bug414855'\n"
			+ "{\n"
			+ "    datatype MyString : 'java.lang.String' { serializable };\n"
			+ "    class ClassExtension extends pivot::Class {}\n"
			+ "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, testFileStem, "platform:/plugin/org.eclipse.ocl.examples.pivot/model/Pivot.merged.genmodel#//pivot");
		doDelete(testProjectName);
		URI fileURI = createModels(testProjectPath, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, fileURI);
		doCompile(testProjectName);		
	}
	
	public void testBug415782() throws Exception {
		String testFileStem = "Bug415782";
		String testProjectName = "bug415782";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile = 
			  "import ecore : 'http://www.eclipse.org/emf/2002/Ecore#/';\n"
			+ "package bug415782 : bug415782 = 'http://bug415782'\n"
			+ "{\n"		
			+ "    class MyClass\n"
			+ "    {\n"
			+ "    	   attribute manyDates : ecore::EDate[*] { ordered };\n"
			+ "        attribute aBool : Boolean;\n"
			+ "        operation anOp() : MyClass {"
			+ "             body : MyClass {"
			+ "               aBool = manyDates->isEmpty()\n"
			+ "             };"
			+ "        }\n"
			+ "    }\n"
			+ "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, testFileStem, null);
		doDelete(testProjectName);
		URI fileURI = createModels(testProjectPath, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, fileURI);
		doCompile(testProjectName);
	}
	
	public void testBug416421() throws Exception {
		String testFileStemA = "Bug416421A";
		String testProjectNameA = "bug416421A";
		String testProjectPathA = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectNameA : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFileA = 
				  "package bug416421A : bug416421A = 'example.org/bug416421A'\n"
				+ "{\n"
				+ "	class ClassA\n"
				+ "	{\n"
				+ "		operation getFalse() : Boolean\n"
				+ "		{\n"
				+ "			body: false;\n"
				+ "		}\n"
				+ "	}\n"
				+ "}\n";
		String genmodelFileA = createGenModelContent(testProjectPathA, testFileStemA, null);
		String testFileStemB = "Bug416421B";
		String testProjectNameB = "bug416421B";
		String testProjectPathB = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectNameB : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFileB = 
				  "import bug416421A : 'Bug416421A.ecore#/';\n"
				+ "package bug416421B : bug416421B = 'example.org/bug416421B'\n"
				+ "{\n"
				+ "	class ClassB extends bug416421A::ClassA\n"
				+ "	{\n"
				+ "		operation getTrue() : Boolean\n"
				+ "		{\n"
				+ "			body: true;\n"
				+ "		}\n"
				+ "	}\n"
				+ "}\n";
		String genmodelFileB = createGenModelContent(testProjectPathB, testFileStemB, "Bug416421A.genmodel#//bug416421A");
		doDelete(testProjectNameA);
		doDelete(testProjectNameB);
		URI fileURIA = createModels(testProjectPathA, testFileStemA, oclinecoreFileA, genmodelFileA);
		URI fileURIB = createModels(testProjectPathB, testFileStemB, oclinecoreFileB, genmodelFileB);
		// B first demonstrates the demand load of Bug416421A to fix Bug 416421
		doGenModel(testProjectPathB, fileURIB);
		doGenModel(testProjectPathA, fileURIA);
		doCompile(testProjectNameA);
		doCompile(testProjectNameB, testProjectNameA);
	}
	
	public void testBug458772() throws Exception {
		String testFileStem = "Bug458772";
		String testProjectName = "bug458772";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile = 
			  "import ecore : 'http://www.eclipse.org/emf/2002/Ecore';\n" + 
			  "\n" + 
			  "package bug458772 : bug458772 = 'http://www.example.com/bug458772/rootPackage/2.0'\n" + 
			  "{\n" + 
			  "	package subPackage : subPackage = 'http://www.example.com/bug458772/subPackage/2.0'\n" + 
			  "	{\n" + 
			  "		class SubElement\n" + 
			  "		{\n" + 
			  "			operation op(tokens : String[*] { ordered !unique }) : Boolean\n" + 
			  "			{\n" + 
			  "				body: \n" + 
			  "				\n" + 
			  "				if tokens->at(1) = '1'\n" + 
			  "				then\n" + 
			  "					op2(tokens)\n" + 
			  "			    else\n" + 
			  "			    	true\n" + 
			  "			    endif;\n" + 
			  "			}\n" + 
			  "			operation op2(tokens : String[*] { ordered !unique }) : Boolean\n" + 
			  "			{\n" + 
			  "				body: \n" + 
			  "				true;\n" + 
			  "			}\n" + 
			  "		}\n" + 
			  "	}\n" + 
			  "	abstract class Element\n" + 
			  "	{\n" + 
			  "		attribute name : String = '';\n" + 
			  "	}\n" + 
			  "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, testFileStem, null);
		doDelete(testProjectName);
		URI genModelURI = createModels(testProjectPath, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, genModelURI);
		doCompile(testProjectName);
	}
	
	public void testBug458773() throws Exception {
		String testFileStem = "Bug458773";
		String testProjectName = "bug458773";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile = 
			  "import ecore : 'http://www.eclipse.org/emf/2002/Ecore';\n" + 
			  "\n" + 
			  "package bug458773 : bug458773 = 'http://www.example.com/bug458773/rootPackage/1.0'\n" + 
			  "{\n" + 
			  "    package subPackage : subPackage = 'http://www.example.com/bug458773/subPackage/1.0'\n" + 
			  "    {\n" + 
			  "        class Element extends bug458773::Element\n" + 
			  "        {\n" + 
			  "\n" + 
			  "            /*\n" + 
			  "             * Error also occurs with Bag(OclAny) in signature without\n" + 
			  "{!unique}\n" + 
			  "             */\n" + 
			  "            operation op() : ocl::OclAny[*] { !unique }\n" + 
			  "            {\n" + 
			  "                body: \n" + 
			  "                Bag{};\n" + 
			  "            }\n" + 
			  "        }\n" + 
			  "    }\n" + 
			  "    abstract class Element\n" + 
			  "    {\n" + 
			  "        attribute name : String = '';\n" + 
			  "    }\n" + 
			  "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, testFileStem, null);
		doDelete(testProjectName);
		URI genModelURI = createModels(testProjectPath, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, genModelURI);
		doCompile(testProjectName);
	}
	
	public void testBug458774() throws Exception {
		String testFileStem = "Bug458774";
		String testProjectName = "bug458774";
		String testProjectPath = EMFPlugin.IS_ECLIPSE_RUNNING ? testProjectName : ORG_ECLIPSE_OCL_EXAMPLES_XTEXT_TESTRESULTS;
		String oclinecoreFile = 
			  "import ecore : 'http://www.eclipse.org/emf/2002/Ecore';\n" + 
			  "\n" + 
			  "package bug458774 : bug458774 = 'http://www.example.com/bug458774/rootPackage/2.0'\n" + 
			  "{\n" + 
			  "    class Element\n" + 
			  "    {\n" + 
			  "        attribute name : String = '';\n" + 
			  "        invariant\n" + 
			  "        elementNameNotReservedWord: \n" + 
			  "            let name: String = self.name.toLower() in\n" + 
			  "            name <> 'reserved_1' and\n" + 
			  "            name <> 'reserved_2' and\n" + 
			  "            name <> 'reserved_3' and\n" + 
			  "            name <> 'reserved_4' and\n" + 
			  "            name <> 'reserved_5' and\n" + 
			  "            name <> 'reserved_6' and\n" + 
			  "            name <> 'reserved_7' and\n" + 
			  "            name <> 'reserved_8' and\n" + 
			  "            name <> 'reserved_9' and\n" + 
			  "            name <> 'reserved_10' and\n" + 
			  "            name <> 'reserved_11' and\n" + 
			  "            name <> 'reserved_12' and\n" + 
			  "            name <> 'reserved_13' and\n" + 
			  "            name <> 'reserved_14' and\n" + 
			  "            name <> 'reserved_15' and\n" + 
			  "            name <> 'reserved_16' and\n" + 
			  "            name <> 'reserved_17' and\n" + 
			  "            name <> 'reserved_18' and\n" + 
			  "            name <> 'reserved_19';\n" + 
			  "    }\n" + 
			  "}\n";
		String genmodelFile = createGenModelContent(testProjectPath, testFileStem, null);
		doDelete(testProjectName);
		URI genModelURI = createModels(testProjectPath, testFileStem, oclinecoreFile, genmodelFile);
		doGenModel(testProjectPath, genModelURI);
		doCompile(testProjectName);
	}
}
